/*-
 * Copyright (c) 1999, 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sbuf.c,v 1.9 2001/05/15 15:19:23 vlm Exp $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>

#include "strfunc.h"

#define	INITIAL_SBUF_SIZE	16

sbuf *
sbuf_init() {
	register sbuf *sb;
	sb = (sbuf *)sf_calloc(1, sizeof(sbuf));
	if(!sb)
		return NULL;

	/* Initial fill */
	if(sbuf_clear(sb) == -1) {
		free(sb);
		return NULL;
	};

	return sb;
};

int
sbuf_clear(sbuf *sb) {
	char *c;

	if(!sb)
		return -1;

	c = (char *)sf_realloc(sb->buf, INITIAL_SBUF_SIZE);
	if(c) {
		sb->buf = c;
		sb->size = INITIAL_SBUF_SIZE;
	} else {
		if(!sb->buf || !sb->size)
			return -1;
	}

	sb->len = 0;
	*sb->buf = 0;

	return 0;
};

int
sbuf_zero(sbuf *sb) {
	if(!sb)
		return -1;

	sb->len = 0;

	if(sb->buf)
		*sb->buf = 0;
	else
		sbuf_clear(sb);

	return 0;
};

void
sbuf_free(sbuf *sb) {
	if(!sb)
		return;

	if(sb->buf);
		free(sb->buf);

	free(sb);
};

char *
sbuf_detach(sbuf *sb, size_t *optLength, size_t *optSize) {
	char *buf;

	if(!sb)
		return NULL;

	if(!sb->buf)
		sbuf_clear(sb);

	if(optLength)
		*optLength= sb->len;
	if(optSize)
		*optSize = sb->size;

	buf = sb->buf;

	/* Reinitialize buffer */
	sb->buf = NULL;
	sbuf_clear(sb);

	return buf;
};

#define	sizeround(foo,prevsize)	((foo)+((prevsize)-1))&(~(prevsize-1))

ssize_t
sbuf_extend(sbuf *sb, size_t newsize) {
	char *c;

	if(!sb)
		return -1;

	if(newsize < sb->size)
		return sb->size;

	newsize = sizeround(newsize + 1, sb->size);

	c = (char *)sf_realloc(sb->buf, newsize);
	if(!c)
		return -1;

	sb->buf = c;
	sb->size = newsize;

	/* Clear the rest of the buffer */
	c = sb->buf + sb->len;
	newsize = sb->size - sb->len;
	memset(c, 0, newsize);

	return sb->size;
}

ssize_t
sbuf_add(sbuf *sb, char *str) {
	register size_t size;

	if(!str)
		return -1;

	size = strlen(str);

	if( sbuf_add2(sb, str, size) == -1 )
		return -1;

	return size;
};

ssize_t
sbuf_add2(sbuf *sb, char *data, size_t nbytes) {

	if(!sb || !data || !nbytes)
		return -1;

	if((sb->len + nbytes) >= sb->size) {
		char *p;
		size_t newsize = sizeround((sb->len + nbytes + 1), sb->size);

		p = (char *)sf_realloc(sb->buf, newsize);
		if(!p)
			return -1;

		sb->buf = p;
		sb->size = newsize;
	}

	memcpy(sb->buf + sb->len, data, nbytes);

	sb->len += nbytes;
	sb->buf[sb->len] = 0;

	return sb->len;
};


ssize_t
sbuf_sprintf(sbuf *sb, char *fmt, ...) {
	register size_t avail=0;	/* Available buffer size */
	register int size;
	va_list ap;

	if(!sb || !fmt)
		return -1;

	va_start(ap, fmt);

	if((avail = (sb->size - sb->len)) < 2) {
		if(sbuf_extend(sb, sb->size + 2) == -1)
			return -1;
		avail = sb->size - sb->len;
	};

	while(avail <= (size=vsnprintf(sb->buf + sb->len, avail, fmt, ap))) {
		if(sbuf_extend(sb, sb->size + size) == -1) {
			/* Clear the tails... */
			sb->buf[sb->len] = 0;
			return -1;
		}
		avail = sb->size - sb->len;
	};

	sb->len += size;

	va_end(ap);

	return size;
};

char *
sbuf_fgets(sbuf *sb, FILE *fd) {
	register size_t avail=0;	/* Available buffer size */
	register ssize_t saved=-1;
	register char *p;

	if(!sb || !fd)
		return NULL;

  while(1) {

	if((avail = (sb->size - sb->len)) < 2) {
		if(sbuf_extend(sb, sb->size + 2) == -1)
			return NULL;
		avail = sb->size - sb->len;
	};

	p = fgets(sb->buf + sb->len, avail, fd);
	if(!p) {
		/* Something still in the buffer */
		if(saved != -1)
			return (sb->buf + saved);
		return NULL;
	};

	if(saved == -1)
		saved = sb->len;

	if(!*p)
		continue;

	sb->len += strlen(p);
	if( sb->buf[sb->len - 1] == '\n' )
		return sb->buf + saved;

  };	/* while(1) */

	return NULL;
};


