/*-
 * Copyright (c) 1999, 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: strfunc.c,v 1.9 2001/05/15 15:19:23 vlm Exp $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "strfunc.h"

int eq(const char *a, const char *b) { return !strcmp(a, b); }

int
ecq(const char *a, const char *b) {
	register const unsigned char
		*us1 = (const unsigned char *)a,
		*us2 = (const unsigned char *)b;
		while (tolower(*us1) == tolower(*us2++))
			if (*us1++ == '\0')
				return 1;

		return 0;
}

int
ecq2(const char *a, const char *b) {
	register const unsigned char
		*us1 = (const unsigned char *)a,
		*us2 = (const unsigned char *)b;
		while (tolower(*us1) == *us2++)
			if (*us1++ == '\0')
				return (1);

	return 0;
};

/* create or initialize empty slist structure */
slist *
sinit() {
	slist *sl;

	sl=(slist *)sf_calloc(1, sizeof(slist));
	sl->listlen = 4;

	sl->list = (char **)sf_malloc(sizeof(char *) * sl->listlen);
	sl->lens = (size_t *)sf_malloc(sizeof(size_t) * sl->listlen);

	*(sl->list) = NULL;
	*(sl->lens) = 0;

	return sl;
};

slist *
scopy(slist *sl) {
	slist *sd;
	size_t n;

	if(!sl)
		return NULL;

	sd = sinit();
	if(!sd)
		return NULL;

	if(!sl->count || !sl->list)
		return sd;

	for(n = 0; n < sl->count; n++)
		sadd2(sd, sl->list[n], sl->lens[n]);

	return sd;
};

/* Clears **vp! */
int
simport(slist *sl, char **vp) {
	char **p;
	if(!sl || !vp) {
		if(vp)
			free_values(vp);
		return -1;
	};

	for(p = vp; *p; p++)
		sadd_attach(sl, *p, strlen(*p));

	free(vp);
	return 0;
};

/* frees any strings inside the slist object */
void
sclear(slist *sl) {
	if(!sl)
		return;

	if(sl->listlen > 0) {
		if(sl->list) {
			while(sl->count--)
				if(sl->list[sl->count])
					free(sl->list[sl->count]);
			*(sl->list) = NULL;
			*(sl->lens) = 0;
		} else {
			sl->listlen = 0;
			if(sl->lens) {
				free(sl->lens);
				sl->lens = 0;
			}
		}
	} else {
		sl->listlen = 0;
		if(sl->list) {
			free(sl->list);
			sl->list = 0;
		}
		if(sl->lens) {
			free(sl->lens);
			sl->lens = 0;
		}
	}

	sl->count=0;
	sl->maxlen=0;
}

void
sfree(slist *sl) {
	if(!sl)
		return;

	if(sl->list) {
		if(sl->count > 0)
			while(sl->count--)
				free(sl->list[sl->count]);
		free(sl->list);
	};

	if(sl->lens)
		free(sl->lens);

	free(sl);
};

/* Basic adding function */
int
_sf_add_internal(slist *s, void *msg, size_t len) {

	if(s->count + 1 >= s->listlen) {
		void *z;
		size_t newlistlen;

		if((newlistlen = s->listlen << 2) == 0)
			newlistlen = 4;

		z = sf_realloc(s->list, sizeof(char *) * newlistlen);
		if(z)
			s->list = (char **)z;
		else
			return -1;

		z = sf_realloc(s->lens, sizeof(size_t) * newlistlen);
		if(z)
			s->lens = (size_t *)z;
		else
			return -1;
		s->listlen = newlistlen;
	}

	s->list[s->count] = msg;
	s->lens[s->count] = len;

	if(len > s->maxlen)
		s->maxlen = len;

	s->count++;
	s->list[s->count] = NULL;
	s->lens[s->count] = 0;

	return 0;
}


/* Attach previously malloc'ed memory */
int
sadd_attach(slist *s, char *msg, size_t len) {

	if(!s || !msg) {
		if(msg)
			free(msg);
		return -1;
	}

	msg = (char *)sf_realloc(msg, len + 1);
	msg[len] = 0;

	if(_sf_add_internal(s, msg, len) == -1) {
		free(msg);
		return -1;
	}

	return 0;
}

/* Add a string to the end of the slist structure */
int
sadd2(slist *s, char *msg, size_t len) {
	char *tmp;

	if(!s || !msg)
		return -1;

	tmp = (char *)sf_malloc(len+1);
	memcpy(tmp, msg, len);
	tmp[len] = 0;

	if(_sf_add_internal(s, tmp, len) == -1) {
		free(tmp);
		return -1;
	}

	return 0;
};

int
sadd(slist *s, char *msg) {
	if(!s || !msg)
		return -1;
	return sadd2(s, msg, strlen(msg));
}


/* delete i'th string from the list */
int
sdel(slist *s, size_t n) {
	if(!s)
		return -1;

	if(n > s->count)
		return s->count;

	if(n == s->count-1) {
		free(s->list[n]);
		s->list[n]=NULL;
		return --(s->count);
	}

	free(s->list[n]);

	for(s->count--; n <= s->count; n++) {
		s->list[n] = s->list[n+1];
		s->lens[n] = s->lens[n+1];
	};

	return s->count;
}

/* Insert element */
ssize_t
sins(slist *s, char *el, size_t n) {
	int i;

	if(!s || !el) return -1;
	if(n > s->count) return -1;

	if(sadd(s, el) == -1)
		return -1;

	for(i = s->count; i > n; i--) {
		s->list[i] = s->list[i-1];
		s->lens[i] = s->lens[i-1];
	};
	s->list[n] = s->list[s->count];
	s->lens[n] = s->lens[s->count];

	s->list[s->count] = NULL;
	s->lens[s->count] = NULL;

	return n;
};

ssize_t
sfind(slist *sl, const char *what) {
	int n;
	int wlen;
	short sw;

	if(!sl || !sl->count || !what)
		return -1;

	wlen=strlen(what);
	if(wlen)
		sw = *(short *)what;
	else
		sw = 0;	/* will never be used */

	for(n=0; n < sl->count; n++) {

		if(sl->lens[n] != wlen)
			continue;

		if(wlen) {
			if(
				(*(short *)sl->list[n] == sw)
				&& !strcmp(sl->list[n], what)
			) return n;
		} else {
			return n;
		};

	};

	return -1;
}

ssize_t
scfind(slist *sl, const char *what) {
	size_t n;
	int wlen;

	if(!sl || !sl->count || !what)
		return -1;

	wlen=strlen(what);

	for(n=0; n < sl->count; n++) {

		if(sl->lens[n] != wlen)
			continue;

		if(wlen) {
			if(ecq(sl->list[n], what))
				return n;
		} else {
			return n;
		};
	};

	return -1;
}

char *
sget2(slist *sl, const char *what, slist *sl2) {
	int n;

	if((n=sfind(sl, what)) == -1)
		return NULL;

	if(n >= sl2->count)
		return NULL;

	return sl2->list[n];
}

char *
scget2(slist *sl, const char *what, slist *sl2) {
	int n;

	if((n=scfind(sl, what)) == -1)
		return NULL;

	if(n >= sl2->count)
		return NULL;

	return sl2->list[n];
};


char **
sarray(slist *sl, size_t start) {

	if(!sl)
		return NULL;

	if(start > sl->count)
		start=sl->count;

	return &(sl->list[start]);
}

char **
mkarray(slist *sl, size_t start) {
	size_t n;
	char **v;
	char **vp;

	if(!sl)
		return NULL;

	n=sizeof(char *) * (sl->count - start + 1);
	vp=v=(char **)sf_malloc(n);

	for( n=start; n < sl->count; n++) {
		if(!(*vp=sf_strdup(sl->list[n]))) {
			free_values(v);
			return NULL;
		};
		vp++;
	}

	*vp=NULL;

	return v;
};


/* ==================================================================== */

ssize_t
find(char **vp, const char *what) {
	int z=0;

	if(!vp || !what)
		return -1;

	for(; *vp; vp++, z++)
		if(!strcmp(*vp, what))
			return z;

	return -1;
}

ssize_t
cfind(char **vp, const char *what) {
	int z=0;

	if(!vp || !what)
		return -1;

	for(; *vp; vp++, z++)
		if(ecq(*vp, what))
			return z;

	return -1;
}


